import sys, math, numpy
from random import random
from collections import Counter
import scipy
from scipy.sparse import lil_matrix, linalg

minimum_frequency = 5
window = 2

filename = sys.argv[1]

word_counts = Counter()
pair_counts = Counter()

## Count words. Each line in the file is a document
total_words = 0
for line in open(sys.argv[1]):
    line = line.replace(".", " ").replace(",", " ")
    tokens = line.split()
    total_words += len(tokens)
    for token in tokens:
        word_counts[token] += 1

## Use these to map from ids to strings and back
vocab = []
vocab_ids = dict()

## Build pruned vocabulary 
log_word_counts = {}
for word in word_counts.keys():
    if word_counts[word] >= minimum_frequency and len(word) > 1:
        vocab_ids[word] = len(vocab)
        vocab.append(word)
        log_word_counts[word] = math.log(word_counts[word])

## Build word2vec token sub-sampling distribution
sampling_distribution = {}
for word in word_counts.keys():
    score = float(word_counts[word]) / (0.0001 * total_words)
    sampling_distribution[word] = (1.0 + math.sqrt(score)) / score

## Run through the docs again to count word pairs
total_pairs = 0
for line in open(sys.argv[1]):
    line = line.replace(".", " ").replace(",", " ")
    tokens = line.split()
    #tokens = filter(lambda w: random() < sampling_distribution[w], tokens)
    tokens = list(filter(lambda w: w in vocab_ids, tokens ))
    
    for i in range(len(tokens) - window):
        for offset in range(0, window):
            pair = [tokens[i], tokens[i + (offset + 1)]]
            pair.sort()
            pair_counts[" ".join(pair)] += window - offset
            total_pairs += window - offset

## Build matrix representations
sparse_pmi = lil_matrix( (len(vocab), len(vocab)) )

log_offset = 2 * math.log(total_words) - math.log(total_pairs)

for pair in pair_counts.keys():
    (left, right) = pair.split()
    score = math.log(pair_counts[pair]) - log_word_counts[left] - log_word_counts[right] + log_offset
    
    if (score > 0.0):
        sparse_pmi[vocab_ids[left], vocab_ids[right]] = score
        sparse_pmi[vocab_ids[right], vocab_ids[left]] = score
    
    ## Uncomment this to show PMI scores
    #print("{}\t{}\t{}\t{}\t{}\t{}\t{}".format(score, pair_counts[pair], total_pairs, word_counts[left], word_counts[right], left, right))

sparse_pmi = sparse_pmi.tocsr()

embeddings, singular_values, context_embeddings = scipy.sparse.linalg.svds(sparse_pmi, 50)

def scale_by_singular_values(matrix):
    return matrix * numpy.sqrt(singular_values[numpy.newaxis, :])

def l2_norm(matrix):
    row_norms = numpy.sqrt(numpy.sum(matrix ** 2, axis = 1))
    return matrix / row_norms[:, numpy.newaxis]

#embeddings = l2_norm(scale_by_singular_values(embeddings))
embeddings = l2_norm(embeddings)

def print_sorted(x, labels):
    sorted_labels = sorted(enumerate(x), key=lambda w: -w[1])
    for index, score in sorted_labels[0:20]:
        label = labels[index]
        print (index, score, label)

def find_closest_terms(term):
    source_id = vocab_ids[term]
    
    source_vector = embeddings[source_id,:]
    sums = embeddings.dot(source_vector)
    print_sorted(sums, vocab)

